import io

from docutils import nodes
from docutils import writers
from docutils.core import publish_file
from docutils.core import publish_string

from .docutils import setup_docutils
from .environment import DefaultEnvironment
from .environment import Environment


class Writer(writers.Writer):

    supported = ("markdown",)

    def __init__(self, limit_version=None, receive_sections=None):
        super(Writer, self).__init__()
        self.limit_version = limit_version
        self.receive_sections = receive_sections

    def translate(self):
        translator = MarkdownTranslator(
            self.document, self.limit_version, self.receive_sections
        )
        self.document.walkabout(translator)
        self.output = translator.output_buf.getvalue()


class MarkdownTranslator(nodes.NodeVisitor):
    def __init__(self, document, limit_version, receive_sections):
        super(MarkdownTranslator, self).__init__(document)
        self.buf = self.output_buf = io.StringIO()
        self.limit_version = limit_version
        self.receive_sections = receive_sections
        self.section_level = 1
        self.stack = []

        self._standalone_section_display = (
            self.limit_version or self.receive_sections
        )
        if self._standalone_section_display:
            self.disable_writing()

    def enable_writing(self):
        self.buf = self.output_buf

    def disable_writing(self):
        self.buf = io.StringIO()

    def _detect_section_was_squashed_into_subtitle(self, document_node):
        # docutils converts a single section we generated into a "subtitle"
        # and squashes the section object
        # http://docutils.sourceforge.net/FAQ.html\
        # #how-can-i-indicate-the-document-title-subtitle
        for subnode in document_node:
            if "version_string" in subnode.attributes:
                if isinstance(subnode, nodes.subtitle):
                    return subnode
                elif isinstance(subnode, nodes.section):
                    return False
                else:
                    raise NotImplementedError(
                        "detected version_string in unexpected node type: %s"
                        % subnode
                    )

    def visit_document(self, node):
        self.document = node
        self.env = Environment.from_document_settings(self.document.settings)

        subtitle_node = self._detect_section_was_squashed_into_subtitle(node)
        if subtitle_node:
            version = subtitle_node.attributes["version_string"]
            rebuild_our_lost_section = nodes.section(
                "",
                nodes.title(version, version, classes=["release-version"]),
                **subtitle_node.attributes
            )

            # note we are taking the nodes from the document and moving them
            # into the new section.   nodes have a "parent" which means that
            # parent changes, which means we are mutating the internal
            # state of the document node.  so use deepcopy() to avoid this
            # side effect; deepcopy() for these nodes seems to not be a
            # performance problem.
            rebuild_our_lost_section.extend([n.deepcopy() for n in node[2:]])
            rebuild_our_lost_section.walkabout(self)
            raise nodes.SkipNode()

    def visit_standalone_version_node(self, node, version_string):
        """visit a section or document that has a changelog version string
        at the top"""

        if self.limit_version and self.limit_version != version_string:
            return

        self.section_level = 1
        self.enable_writing()
        if self.receive_sections:
            self.buf = io.StringIO()

    def depart_standalone_version_node(self, node, version_string):
        """depart a section or document that has a changelog version string
        at the top"""

        if self.limit_version and self.limit_version != version_string:
            return

        if self.receive_sections:
            self.receive_sections(version_string, self.buf.getvalue())
        self.disable_writing()

    def visit_section(self, node):
        if (
            "version_string" in node.attributes
            and self._standalone_section_display
        ):
            self.visit_standalone_version_node(
                node, node.attributes["version_string"]
            )
        else:
            self.section_level += 1

    def depart_section(self, node):
        if (
            "version_string" in node.attributes
            and self._standalone_section_display
        ):
            self.depart_standalone_version_node(
                node, node.attributes["version_string"]
            )
        else:
            self.section_level -= 1

    def visit_strong(self, node):
        self.buf.write("**")

    def depart_strong(self, node):
        self.buf.write("**")

    def visit_emphasis(self, node):
        self.buf.write("*")

    def depart_emphasis(self, node):
        self.buf.write("*")

    def visit_literal(self, node):
        self.buf.write("`")

    def visit_Text(self, node):
        self.buf.write(node.astext())

    def depart_Text(self, node):
        pass

    def depart_paragraph(self, node):
        self.buf.write("\n\n")

    def depart_literal(self, node):
        self.buf.write("`")

    def visit_title(self, node):
        self.buf.write(
            "\n%s %s\n\n" % ("#" * self.section_level, node.astext())
        )
        raise nodes.SkipNode()

    def visit_changeset_link(self, node):
        # it would be nice to have an absolutely link to the HTML
        # hosted changelog but this requires being able to generate
        # the absolute link from the document filename and all that.
        # it can perhaps be sent on the commandline
        raise nodes.SkipNode()

    def depart_changeset_link(self, node):
        pass

    def visit_reference(self, node):
        if "changelog-reference" in node.attributes["classes"]:
            self.visit_changeset_link(node)
        else:
            self.buf.write("[")

    def depart_reference(self, node):
        if "changelog-reference" in node.attributes["classes"]:
            self.depart_changeset_link(node)
        else:
            self.buf.write("](%s)" % node.attributes["refuri"])

    def visit_admonition(self, node):
        # "seealsos" typically have internal sphinx references so at the
        # moment we're not prepared to look those up, future version can
        # perhaps use sphinx object lookup
        raise nodes.SkipNode()

    def visit_list_item(self, node):
        self.stack.append(self.buf)
        self.buf = io.StringIO()

    def depart_list_item(self, node):
        popped = self.buf
        self.buf = self.stack.pop(-1)

        indent_level = len(self.stack)
        indent_string = " " * 4 * indent_level

        value = popped.getvalue().strip()

        lines = value.split("\n")
        self.buf.write("\n" + indent_string + "-   ")
        line = lines.pop(0)
        self.buf.write(line + "\n")
        for line in lines:
            self.buf.write(indent_string + "    " + line + "\n")

    def _visit_generic_node(self, node):
        pass

    def __getattr__(self, name):
        if not name.startswith("_"):
            return self._visit_generic_node
        else:
            raise AttributeError(name)


def stream_changelog_sections(
    target_filename, config_filename, receive_sections, version=None
):
    """Send individual changelog sections to a callable, one per version.

    The callable accepts two arguments, the string version number of the
    changelog section, and the markdown-formatted content of the changelog
    section.

    Used for APIs that receive changelog sections per version.

    """
    Environment.register(DefaultEnvironment)

    setup_docutils()
    with open(target_filename, encoding="utf-8") as handle:
        publish_string(
            handle.read(),
            source_path=target_filename,
            writer=Writer(
                limit_version=version, receive_sections=receive_sections
            ),
            settings_overrides={
                "changelog_env": DefaultEnvironment(config_filename),
                "report_level": 3,
            },
        )


def render_changelog_as_md(
    target_filename, config_filename, version, sections_only
):

    Environment.register(DefaultEnvironment)

    setup_docutils()

    if sections_only:

        def receive_sections(version_string, text):
            print(text)

    else:
        receive_sections = None

    writer = Writer(limit_version=version, receive_sections=receive_sections)
    settings_overrides = {
        "changelog_env": DefaultEnvironment(config_filename),
        "report_level": 3,
    }

    with open(target_filename, encoding="utf-8") as handle:
        if receive_sections:
            publish_string(
                handle.read(),
                source_path=target_filename,
                writer=writer,
                settings_overrides=settings_overrides,
            )
        else:
            publish_file(
                handle, writer=writer, settings_overrides=settings_overrides
            )



import os

from sphinx.util import logging
from sphinx.util import status_iterator
from sphinx.util.console import bold
from sphinx.util.osutil import copyfile

from .docutils import ChangeDirective
from .docutils import ChangeLogDirective
from .docutils import ChangeLogImportDirective
from .docutils import make_ticket_link
from .environment import Environment


LOG = logging.getLogger(__name__)


def _is_html(app):
    return app.builder.name in ("html", "readthedocs")


class SphinxEnvironment(Environment):
    __slots__ = ("sphinx_env",)

    @classmethod
    def from_document_settings(cls, settings):
        return SphinxEnvironment(settings.env)

    def __init__(self, sphinx_env):
        self.sphinx_env = sphinx_env

    def log_debug(self, msg, *args):
        LOG.debug(msg, *args)

    @property
    def temp_data(self):
        return self.sphinx_env.temp_data

    @property
    def changelog_sections(self):
        return self.sphinx_env.config.changelog_sections

    @property
    def changelog_caption_class(self):
        return self.sphinx_env.config.changelog_caption_class

    @property
    def changelog_inner_tag_sort(self):
        return self.sphinx_env.config.changelog_inner_tag_sort

    @property
    def changelog_hide_sections_from_tags(self):
        return self.sphinx_env.config.changelog_hide_sections_from_tags

    @property
    def changelog_hide_tags_in_entry(self):
        return self.sphinx_env.config.changelog_hide_tags_in_entry

    @property
    def changelog_render_ticket(self):
        return self.sphinx_env.config.changelog_render_ticket

    @property
    def changelog_render_pullreq(self):
        return self.sphinx_env.config.changelog_render_pullreq

    @property
    def changelog_render_changeset(self):
        return self.sphinx_env.config.changelog_render_changeset

    def status_iterator(self, elements, message):
        return status_iterator(
            elements,
            message,
            "purple",
            length=len(elements),
            verbosity=self.sphinx_env.app.verbosity,
        )


def add_stylesheet(app):
    # changed in 1.8 from add_stylesheet()
    # https://www.sphinx-doc.org/en/master/extdev/appapi.html#sphinx.application.Sphinx.add_css_file
    app.add_css_file("changelog.css")


def copy_stylesheet(app, exception):
    LOG.info(
        bold("The name of the builder is: %s" % app.builder.name), nonl=True
    )

    if not _is_html(app) or exception:
        return
    LOG.info(bold("Copying sphinx_paramlinks stylesheet... "), nonl=True)

    source = os.path.abspath(os.path.dirname(__file__))

    # the '_static' directory name is hardcoded in
    # sphinx.builders.html.StandaloneHTMLBuilder.copy_static_files.
    # would be nice if Sphinx could improve the API here so that we just
    # give it the path to a .css file and it does the right thing.
    dest = os.path.join(app.builder.outdir, "_static", "changelog.css")
    copyfile(os.path.join(source, "changelog.css"), dest)
    LOG.info("done")


def setup(app):
    Environment.register(SphinxEnvironment)
    app.add_directive("changelog", ChangeLogDirective)
    app.add_directive("change", ChangeDirective)
    app.add_directive("changelog_imports", ChangeLogImportDirective)
    app.add_config_value("changelog_sections", [], "env")
    app.add_config_value("changelog_caption_class", "caption", "env")
    app.add_config_value("changelog_inner_tag_sort", [], "env")
    app.add_config_value("changelog_hide_sections_from_tags", False, "env")
    app.add_config_value("changelog_hide_tags_in_entry", False, "env")
    app.add_config_value("changelog_render_ticket", None, "env")
    app.add_config_value("changelog_render_pullreq", None, "env")
    app.add_config_value("changelog_render_changeset", None, "env")
    app.connect("builder-inited", add_stylesheet)
    app.connect("build-finished", copy_stylesheet)
    app.add_role("ticket", make_ticket_link)

    return {"parallel_read_safe": True, "parallel_write_safe": True}



import argparse
import os
import re
import shutil
import sys
import tempfile

from . import mdwriter


def release_notes_into_changelog_file(target_filename, version, release_date):
    """Read changelog fragment files and render them into a single .rst file.

    remove the fragment files afterwards using git rm.

    The fragment files are located by looking for ':include_notes_from:'
    directives in the given changelog file.

    """
    output = tempfile.NamedTemporaryFile(
        mode="w", delete=False, encoding="utf-8"
    )
    with open(target_filename) as handle:
        for line in handle:
            m = re.match(r".*:version: %s" % version, line)
            if m:
                output.write(line)
                output.write("    :released: %s\n" % release_date)
                continue

            m = re.match(r".*:include_notes_from: (.+)", line)
            if m:
                notes_dir = os.path.join(
                    os.path.dirname(target_filename), m.group(1)
                )
                for fname in os.listdir(notes_dir):
                    if not fname.endswith(".rst"):
                        continue
                    fname_path = os.path.join(notes_dir, fname)
                    output.write("\n")
                    with open(fname_path) as inner:
                        for inner_line in inner:
                            output.write(("    " + inner_line).rstrip() + "\n")
                    os.system("git rm %s" % fname_path)
            else:
                output.write(line)
    output.close()
    shutil.move(output.name, target_filename)


def main(argv=None):
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()

    subparser = subparsers.add_parser(
        "release-notes", help="Merge notes files into changelog and git rm"
    )
    subparser.add_argument("filename", help="target changelog filename")
    subparser.add_argument(
        "version", help="version string as it appears in changelog"
    )
    subparser.add_argument("date", help="full text of datestamp to insert")
    subparser.set_defaults(
        cmd=(
            release_notes_into_changelog_file,
            ["filename", "version", "date"],
        )
    )

    subparser = subparsers.add_parser(
        "generate-md", help="Generate file into markdown"
    )
    subparser.add_argument("filename", help="target changelog filename")
    subparser.add_argument("-c", "--config", help="path to conf.py")
    subparser.add_argument(
        "-v",
        "--version",
        type=str,
        help="render changelog only for version given",
    )
    subparser.add_argument(
        "-s",
        "--sections-only",
        action="store_true",
        help="render changelogs as top level sections",
    )
    subparser.set_defaults(
        cmd=(
            mdwriter.render_changelog_as_md,
            ["filename", "config", "version", "sections_only"],
        )
    )

    options = parser.parse_args(argv)
    fn, argnames = options.cmd
    fn(*[getattr(options, name) for name in argnames])


if __name__ == "__main__":
    main(sys.argv)



#! coding: utf-8
import collections
import itertools

from docutils import nodes


def render_changelog(changelog_directive):
    changes = changelog_directive.get_changes_list(
        changelog_directive.env, changelog_directive.version
    ).values()
    output = []

    id_prefix = "change-%s" % (changelog_directive.version,)
    topsection = _run_top(changelog_directive, id_prefix)
    output.append(topsection)

    bysection, all_sections = _organize_by_section(
        changelog_directive, changes
    )

    counter = itertools.count()

    sections_to_render = [
        s for s in changelog_directive.sections if s in all_sections
    ]
    if not sections_to_render:
        for cat in changelog_directive.inner_tag_sort:
            append_sec = _append_node(changelog_directive)

            for rec in bysection[(changelog_directive.default_section, cat)]:
                rec["id"] = "%s-%s" % (id_prefix, next(counter))

                _render_rec(changelog_directive, rec, None, cat, append_sec)

            if append_sec.children:
                topsection.append(append_sec)
    else:
        for section in sections_to_render + [
            changelog_directive.default_section
        ]:
            sec = nodes.section(
                "",
                nodes.title(section, section),
                ids=["%s-%s" % (id_prefix, section.replace(" ", "-"))],
            )

            append_sec = _append_node(changelog_directive)
            sec.append(append_sec)

            for cat in changelog_directive.inner_tag_sort:
                for rec in bysection[(section, cat)]:
                    rec["id"] = "%s-%s" % (id_prefix, next(counter))
                    _render_rec(
                        changelog_directive, rec, section, cat, append_sec
                    )

            if append_sec.children:
                topsection.append(sec)

    return output


def _organize_by_section(changelog_directive, changes):
    compound_sections = [
        (s, s.split(" ")) for s in changelog_directive.sections if " " in s
    ]

    bysection = collections.defaultdict(list)
    all_sections = set()
    for rec in changes:
        assert changelog_directive.version == rec["render_for_version"]

        inner_tag = rec["tags"].intersection(
            changelog_directive.inner_tag_sort
        )
        if inner_tag:
            inner_tag = inner_tag.pop()
        else:
            inner_tag = ""

        for compound, comp_words in compound_sections:
            if rec["tags"].issuperset(comp_words):
                bysection[(compound, inner_tag)].append(rec)
                all_sections.add(compound)
                break
        else:
            intersect = rec["tags"].intersection(changelog_directive.sections)
            if intersect:
                for sec in rec["sorted_tags"]:
                    if sec in intersect:
                        bysection[(sec, inner_tag)].append(rec)
                        all_sections.add(sec)
                        break
            else:
                bysection[
                    (changelog_directive.default_section, inner_tag)
                ].append(rec)
    return bysection, all_sections


def _append_node(changelog_directive):
    return nodes.bullet_list()


def _run_top(changelog_directive, id_prefix):
    version = changelog_directive._parsed_content.get("version", "")
    topsection = nodes.section(
        "",
        nodes.title(version, version, classes=["release-version"]),
        ids=[id_prefix],
        version_string=version,
    )

    if changelog_directive._parsed_content.get("released"):
        topsection.append(
            nodes.Text(
                "Released: %s"
                % changelog_directive._parsed_content["released"]
            )
        )
    else:
        topsection.append(nodes.Text("no release date"))

    intro_para = nodes.paragraph("", "")
    len_ = -1
    for len_, text in enumerate(changelog_directive._parsed_content["text"]):
        if ".. change::" in text:
            break

    # if encountered any text elements that didn't start with
    # ".. change::", those become the intro
    if len_ > 0:
        changelog_directive.state.nested_parse(
            changelog_directive._parsed_content["text"][0:len_], 0, intro_para
        )
        topsection.append(intro_para)

    return topsection


def _render_rec(changelog_directive, rec, section, cat, append_sec):
    para = rec["node"].deepcopy()

    targetid = "change-%s" % (
        rec["version_to_hash"][changelog_directive.version],
    )
    targetnode = nodes.target("", "", ids=[targetid])

    sections = section.split(" ") if section else []
    section_tags = [tag for tag in sections if tag in rec["tags"]]
    category_tags = [cat] if cat in rec["tags"] else []
    other_tags = list(
        sorted(rec["tags"].difference(section_tags + category_tags))
    )

    all_items = []

    if not changelog_directive.hide_sections_from_tags:
        all_items.extend(section_tags)
    all_items.extend(category_tags)
    all_items.extend(other_tags)

    all_items = all_items or ["no_tags"]

    permalink = nodes.reference(
        "",
        "",
        nodes.Text(u"¶", u"¶"),
        refid=targetid,
        classes=["changelog-reference", "headerlink"],
    )

    if not changelog_directive.hide_tags_in_entry:
        tag_node = nodes.strong("", " ".join("[%s]" % t for t in all_items))
        targetnode.insert(0, nodes.Text(" ", " "))
        targetnode.insert(0, tag_node)
        targetnode.append(permalink)
        para.insert(0, targetnode)

    if len(rec["versions"]) > 1:

        backported_changes = rec["sorted_versions"][
            rec["sorted_versions"].index(changelog_directive.version) + 1 :
        ]
        if backported_changes:
            backported = nodes.paragraph("")
            backported.append(nodes.Text("This change is also ", ""))
            backported.append(nodes.strong("", "backported"))
            backported.append(
                nodes.Text(" to: %s" % ", ".join(backported_changes), "")
            )
            para.append(backported)

    if changelog_directive.hide_tags_in_entry:
        para.append(permalink)

    insert_ticket = nodes.paragraph("")
    para.append(insert_ticket)

    i = 0
    for collection, render, prefix in (
        (
            rec["tickets"],
            changelog_directive.env.changelog_render_ticket,
            "#%s",
        ),
        (
            rec["pullreq"],
            changelog_directive.env.changelog_render_pullreq,
            "pull request %s",
        ),
        (
            rec["changeset"],
            changelog_directive.env.changelog_render_changeset,
            "r%s",
        ),
    ):
        for refname in sorted(collection):
            if i > 0:
                insert_ticket.append(nodes.Text(", ", ", "))
            else:
                insert_ticket.append(nodes.Text("References: " ""))
            i += 1
            if render is not None:
                if isinstance(render, dict):
                    if ":" in refname:
                        typ, refval = refname.split(":")
                    else:
                        typ = "default"
                        refval = refname
                    refuri = render[typ] % refval
                else:
                    refuri = render % refname
                node = nodes.reference(
                    "",
                    "",
                    nodes.Text(prefix % refname, prefix % refname),
                    refuri=refuri,
                )
            else:
                node = nodes.Text(prefix % refname, prefix % refname)
            insert_ticket.append(node)

    append_sec.append(
        nodes.list_item("", nodes.target("", "", ids=[rec["id"]]), para)
    )
